
/*  d4zap.c   (c) Copyright Sequiter Software Inc., 1987-1990.  All rights reserved.
*/

#include "p4misc.h"
#include "d4all.h"
#include "u4error.h"

#ifndef UNIX
#include <fcntl.h>
#include <io.h>
#endif


#ifdef NO_CHSIZE

int d4zap( long start_rec, long end_rec )
{
   u4error( 0, "Routine d4zap is not present when compiled with NO_CHSIZE", (char *) 0 ) ;
   return 0 ;
}

#else

extern  int  d4adjust( BASE *, long ) ;
extern  int  v4lock_wait ;

static int  cleanup(void) ;

static int  cleanup()
{
   d4buf_calc() ;
   d4buf_return() ;
   return -1 ;
}

int d4zap( long start_rec, long end_rec )
{
   long   rec_count, rec_width, read_bytes, read_rec, n_zap ;
   long   tot_bytes, ptr_recs ;
   int    rc ;
   BASE  *base_ptr ;
   char   H_PTR  ptr ;

   if ( (rc = d4lock_all( v4lock_wait, 1)) < 0)  return rc ;

   base_ptr =  d4ptr() ;
   rec_count =  d4reccount() ;
   rec_width =  f4record_width() ;

   if ( end_rec >  rec_count )
	end_rec =  rec_count ;
   if ( start_rec < 1 )
	start_rec = 1 ;

   if ( start_rec > end_rec )  return 0 ;

   if ( d4buf_clear_base(d4select(-1)) < 0 )  return -1 ;
   if ( d4buf_clear_lend(1) == -1 )   return -1 ;
   d4buf_avail_set() ;

   ptr =  (char H_PTR) d4buf_alloc( tot_bytes= d4buf_avail(), 1 ) ;
   if ( tot_bytes < rec_width )
   {
      tot_bytes = rec_width*2 ;
      ptr = (char H_PTR) base_ptr->buffer ;
   }
   ptr_recs  =  tot_bytes/ rec_width ;
   tot_bytes =  ptr_recs * rec_width ;

   n_zap =  end_rec - start_rec +1L ;

   for ( read_rec = end_rec+1L; read_rec <= rec_count; read_rec += ptr_recs)
   {
      if ( d4lseek(read_rec) < 0 )  return( cleanup() ) ;

      read_bytes =  u4huge_read( base_ptr->file_hand, ptr, tot_bytes ) ;
      if ( read_bytes < 0 )  return( cleanup() ) ;

      if ( read_bytes < tot_bytes )
      {
         ptr_recs =  read_bytes/rec_width ;
	 if ( ptr_recs == 0L )
	 {
	    u4error( E_READ, d4name(), (char *) 0 ) ;
	    return( cleanup() ) ;
         }
      }

      if ( d4lseek( read_rec-n_zap) < 0 )  return( cleanup() ) ;
      if ( u4huge_write( base_ptr->file_hand, ptr, read_bytes) < 0 )
         return( cleanup() ) ;
   }

   cleanup() ;

   return ( d4adjust(base_ptr, rec_count-n_zap) ) ;
}

#endif
